import { InjectionToken, Inject, Injectable, Injector, Pipe, isDevMode, NgModule, Optional } from '@angular/core';
import 'rxjs/add/operator/toPromise';
import { ReduxActionDispatcher, ReduxStateDecorator, ReduxReducerDecorator, ReduxStateDecoratorForClass } from '@harmowatch/redux-decorators';export { ReduxActionContextDecoratorForClass as ReduxActionContext, ReduxActionDecoratorForMethod as ReduxAction, ReduxReducerDecoratorForMethod as ReduxReducer, ReduxStateDecoratorForClass as ReduxState } from '@harmowatch/redux-decorators';
import { AsyncSubject } from 'rxjs/AsyncSubject';
import { Observable } from 'rxjs/Observable';
import 'rxjs/add/operator/map';
import { ReplaySubject } from 'rxjs/ReplaySubject';
import 'rxjs/add/operator/distinctUntilChanged';
import 'rxjs/add/operator/takeWhile';
import { BehaviorSubject } from 'rxjs/BehaviorSubject';
import { __decorate } from 'tslib';
import { createStore } from 'redux';
import { CommonModule } from '@angular/common';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
class ReduxStateDefinitionToken extends InjectionToken {
    constructor() {
        super('ReduxStateDefinitionToken');
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
class ReduxStore extends InjectionToken {
    constructor() {
        super('ReduxStore');
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
/**
 * @record
 */

class ReduxRegistry {
    /**
     * @param {?=} store
     */
    constructor(store = null) {
        ReduxRegistry.reset();
        ReduxRegistry.registerStore(store);
    }
    /**
     * @return {?}
     */
    static reset() {
        ReduxRegistry._store = new AsyncSubject();
    }
    /**
     * @param {?} store
     * @return {?}
     */
    static registerStore(store) {
        ReduxRegistry.reset();
        ReduxRegistry._store.next(store);
        ReduxRegistry._store.complete();
        ReduxActionDispatcher.dispatchedActions.subscribe(action => {
            const /** @type {?} */ reduxAction = {
                type: action.type,
                payload: action.payload,
            };
            store.dispatch(reduxAction);
            if (action.onDispatchSuccess) {
                action.onDispatchSuccess();
            }
        });
    }
    /**
     * @param {?} state
     * @return {?}
     */
    static registerState(state) {
        ReduxRegistry.getStore().then((store) => {
            const /** @type {?} */ stateConfig = ReduxStateDecorator.get(state.constructor);
            const /** @type {?} */ initialState = state.getInitialState();
            Promise
                .resolve(initialState instanceof Observable ? initialState.toPromise() : initialState)
                .then(initialValue => {
                store.dispatch({
                    payload: {
                        initialValue,
                        name: stateConfig.name,
                    },
                    type: ReduxRegistry.ACTION_REGISTER_STATE,
                });
            });
        });
    }
    /**
     * @return {?}
     */
    static getStore() {
        return new Promise(ReduxRegistry._store.subscribe.bind(ReduxRegistry._store));
    }
}
ReduxRegistry.ACTION_REGISTER_STATE = `@harmowatch/ngx-redux-core/registerState`;
ReduxRegistry._store = new AsyncSubject();
ReduxRegistry.decorators = [
    { type: Injectable },
];
/** @nocollapse */
ReduxRegistry.ctorParameters = () => [
    { type: undefined, decorators: [{ type: Inject, args: [ReduxStore,] },] },
];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
class ReduxSelector extends ReplaySubject {
    /**
     * @param {?=} selector
     * @param {?=} stateProvider
     */
    constructor(selector = '/', stateProvider) {
        if (!selector.startsWith(ReduxSelector.DELIMITER) && !stateProvider) {
            throw new Error('You need to provide a state provider, if you use relative selectors');
        }
        super(1);
        ReduxRegistry.getStore().then(store => {
            const /** @type {?} */ next = () => {
                this.next(ReduxSelector.getValueByState(store.getState(), selector, stateProvider));
            };
            store.subscribe(() => next());
            next(); // we need to trigger a initial value, otherwise we've to wait until the first state change
        });
    }
    /**
     * @param {?} selector
     * @param {?=} stateProvider
     * @return {?}
     */
    static normalize(selector, stateProvider) {
        if (!selector.startsWith(ReduxSelector.DELIMITER)) {
            const /** @type {?} */ stateName = ReduxStateDecorator.get(stateProvider).name;
            return `${ReduxSelector.DELIMITER}${stateName}${ReduxSelector.DELIMITER}${selector}`;
        }
        return selector;
    }
    /**
     * @template S
     * @param {?} state
     * @param {?} selector
     * @param {?=} stateProvider
     * @return {?}
     */
    static getValueByState(state, selector, stateProvider) {
        /* save the return value in a constant to prevent
             * "Metadata collected contains an error that will be reported at runtime: Lambda not supported."
             * error
             */
        const /** @type {?} */ value = ReduxSelector.normalize(selector, stateProvider).split(ReduxSelector.DELIMITER)
            .filter(propertyKey => propertyKey !== '')
            .reduce((previousValue, propertyKey) => {
            if (!previousValue || !previousValue.hasOwnProperty(propertyKey)) {
                return null;
            }
            return previousValue[propertyKey];
        }, /** @type {?} */ (state));
        return value;
    }
}
ReduxSelector.DELIMITER = '/';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
/**
 * @abstract
 */
class ReduxStateProvider {
    /**
     * @param {?=} stateDefs
     */
    constructor(stateDefs = []) {
        this.selectorCache = {};
        const { name = null } = ReduxStateDecorator.get(this.constructor) || {};
        if (!name) {
            throw new Error('Unable to resolve state name! Make sure you\'ve decorated the provider by "@ReduxState"!');
        }
        this.name = ReduxStateDecorator.get(this.constructor).name;
        this.stateDef = stateDefs.find(def => ReduxStateDecorator.get(def.provider).name === name);
        if (!this.stateDef) {
            throw new Error('Unable to resolve state definition! Make sure you\'ve registered the provider to ReduxModule!');
        }
        this.reducerMethodsByType = (this.stateDef.reducers || [])
            .map(clazz => this.getReducerMethods(new clazz()))
            .reduce((all, curr) => [].concat(curr, all), []) // [].concat keeps the order, all.concat(curr) destroys the order
            .reduce((methodsByType, reducer) => {
            const /** @type {?} */ type = ReduxActionDispatcher.getType(reducer.type);
            return Object.assign({}, methodsByType, { [type]: [reducer.method].concat(methodsByType[type] || []) });
        }, {});
        ReduxStateProvider.instancesByName[this.name] = this;
    }
    /**
     * @return {?}
     */
    getInitialState() {
        throw new Error('Method "getInitialState" not implemented.');
    }
    /**
     * @template T
     * @param {?=} selector
     * @return {?}
     */
    select(selector = '') {
        const /** @type {?} */ stateType = /** @type {?} */ (this.constructor);
        selector = ReduxSelector.normalize(selector, stateType);
        if (!this.selectorCache[selector]) {
            this.selectorCache[selector] = new ReduxSelector(selector, stateType).distinctUntilChanged();
        }
        return /** @type {?} */ (this.selectorCache[selector]);
    }
    /**
     * @return {?}
     */
    getState() {
        return ReduxRegistry.getStore().then(store => {
            return ReduxSelector.getValueByState(store.getState(), '/' + this.name);
        });
    }
    /**
     * @template P
     * @param {?} state
     * @param {?} action
     * @return {?}
     */
    reduce(state, action) {
        const /** @type {?} */ reducerMethods = this.reducerMethodsByType[action.type] || [];
        return reducerMethods.reduce((stateToReduce, method) => method(stateToReduce, action), state);
    }
    /**
     * @param {?} reducerClassInstance
     * @return {?}
     */
    getReducerMethods(reducerClassInstance) {
        return Object.values(Object.getPrototypeOf(reducerClassInstance))
            .map(method => {
            return {
                method: (/** @type {?} */ (method)).bind(reducerClassInstance),
                type: ReduxReducerDecorator.get(method),
            };
        })
            .filter(type => type != null)
            .reduce((all, curr) => all.concat([].concat(curr.type).map(type => (Object.assign({}, curr, { type })))), []);
    }
}
ReduxStateProvider.instancesByName = {};
/** @nocollapse */
ReduxStateProvider.ctorParameters = () => [
    { type: Array, decorators: [{ type: Inject, args: [ReduxStateDefinitionToken,] },] },
];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
/**
 * @template S
 * @param {?} expression
 * @param {?=} context
 * @return {?}
 */
function ReduxSelect(expression, context) {
    return (target, propertyKey) => {
        const /** @type {?} */ stateName = ReduxStateDecorator.get(context).name;
        Object.defineProperty(target, propertyKey, {
            get: () => ReduxStateProvider.instancesByName[stateName].select(expression)
        });
    };
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
class ReduxReducerProvider {
    constructor() {
        this.stateProviders = {};
    }
    /**
     * @return {?}
     */
    get rootReducer() {
        return this.reduce.bind(this);
    }
    /**
     * @param {?} provider
     * @return {?}
     */
    addStateProvider(provider) {
        if (!this.stateProviders[provider.name]) {
            ReduxRegistry.registerState(provider);
            this.stateProviders[provider.name] = provider;
        }
        else {
            throw new Error(`State "${provider.name}" is registered twice! Make sure your state name is unique!`);
        }
    }
    /**
     * @param {?} rootState
     * @param {?} action
     * @return {?}
     */
    reduce(rootState, action) {
        if (action.type === ReduxRegistry.ACTION_REGISTER_STATE) {
            const /** @type {?} */ regAction = (/** @type {?} */ ((action)));
            return Object.assign({}, rootState, { [regAction.payload.name]: regAction.payload.initialValue });
        }
        return Object.values(this.stateProviders).reduce((stateToReduce, provider) => {
            return Object.assign({}, stateToReduce, {
                [provider.name]: provider.reduce(stateToReduce[provider.name], action),
            });
        }, rootState);
    }
}
ReduxReducerProvider.decorators = [
    { type: Injectable },
];
/** @nocollapse */
ReduxReducerProvider.ctorParameters = () => [];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
class ReduxSelectPipe {
    /**
     * @param {?=} stateDefs
     * @param {?=} injector
     */
    constructor(stateDefs = [], injector) {
        this.provider = /** @type {?} */ (injector.get(stateDefs[0].provider));
    }
    /**
     * @param {?} selector
     * @return {?}
     */
    transform(selector) {
        return this.provider.select(selector);
    }
}
ReduxSelectPipe.decorators = [
    { type: Pipe, args: [{ name: 'reduxSelect' },] },
];
/** @nocollapse */
ReduxSelectPipe.ctorParameters = () => [
    { type: Array, decorators: [{ type: Inject, args: [ReduxStateDefinitionToken,] },] },
    { type: Injector, },
];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
class ReduxTestingStore {
    constructor() {
        this.state = new BehaviorSubject(null);
    }
    /**
     * @return {?}
     */
    static factory() {
        return new ReduxTestingStore();
    }
    /**
     * @template S
     * @param {?} state
     * @param {?} value
     * @return {?}
     */
    setState(state, value) {
        const { name } = ReduxStateDecorator.get(state);
        const /** @type {?} */ nextState = Object.assign({}, this.state.getValue(), {
            [name]: value,
        });
        this.state.next(nextState);
        return this.state
            .takeWhile(currentState => currentState !== nextState)
            .toPromise().then(() => this.state.getValue());
    }
    /**
     * @return {?}
     */
    getState() {
        return this.state.getValue();
    }
    /**
     * @param {?} listener
     * @return {?}
     */
    subscribe(listener) {
        return this.state.subscribe.call(this.state, listener);
    }
    /**
     * @return {?}
     */
    replaceReducer() {
    }
    /**
     * @template T
     * @param {?} action
     * @return {?}
     */
    dispatch(action) {
        return action;
    }
}
ReduxTestingStore.decorators = [
    { type: Injectable },
];
/** @nocollapse */
ReduxTestingStore.ctorParameters = () => [];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
/**
 * @record
 */

let TestingStateProvider = TestingStateProvider_1 = class TestingStateProvider extends ReduxStateProvider {
    /**
     * @return {?}
     */
    getInitialState() {
        return TestingStateProvider_1.INITIAL_STATE;
    }
};
TestingStateProvider.NAME = 'testing-7c66b613-20bd-4d35-8611-5181ca4a0b72';
TestingStateProvider.INITIAL_STATE = {
    todo: {
        isFetching: false,
        items: ['Item 1', 'Item 2'],
    },
};
TestingStateProvider = TestingStateProvider_1 = __decorate([
    ReduxStateDecoratorForClass({ name: TestingStateProvider_1.NAME })
], TestingStateProvider);
var TestingStateProvider_1;

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
class ReduxModule {
    /**
     * @param {?} injector
     * @param {?} reducerProvider
     * @param {?=} stateDefs
     */
    constructor(injector, reducerProvider, stateDefs = []) {
        injector.get(ReduxRegistry); // just make sure the provider is instantiated
        if (Array.isArray(stateDefs)) {
            stateDefs
                .filter(def => def && def.provider)
                .map(def => injector.get(def.provider))
                .forEach(provider => reducerProvider.addStateProvider(provider));
        }
    }
    /**
     * @template S
     * @param {?=} config
     * @return {?}
     */
    static forChild(config = {}) {
        return {
            ngModule: ReduxModule,
            providers: [
                { provide: ReduxStateDefinitionToken, useValue: config.state || null, multi: true },
            ],
        };
    }
    /**
     * @template S
     * @param {?=} config
     * @return {?}
     */
    static forRoot(config = {}) {
        return {
            ngModule: ReduxModule,
            providers: [
                ReduxReducerProvider,
                ReduxRegistry,
                {
                    provide: ReduxStore,
                    useFactory: config.storeFactory || ReduxModule.defaultStoreFactory,
                    deps: [ReduxReducerProvider]
                },
                { provide: ReduxStateDefinitionToken, useValue: config.state || null, multi: true },
            ],
        };
    }
    /**
     * @param {?} reduxReducerProvider
     * @param {?=} devMode
     * @return {?}
     */
    static defaultStoreFactory(reduxReducerProvider, devMode = isDevMode()) {
        return createStore(reduxReducerProvider.rootReducer, {}, ReduxModule.defaultEnhancerFactory(devMode));
    }
    /**
     * @param {?} devMode
     * @return {?}
     */
    static defaultEnhancerFactory(devMode) {
        if (window['__REDUX_DEVTOOLS_EXTENSION__'] && devMode) {
            return window['__REDUX_DEVTOOLS_EXTENSION__']();
        }
        return ReduxModule.noopEnhancer;
    }
    /**
     * @param {?} next
     * @return {?}
     */
    static noopEnhancer(next) {
        return next;
    }
}
ReduxModule.decorators = [
    { type: NgModule, args: [{
                declarations: [
                    ReduxSelectPipe,
                ],
                exports: [
                    ReduxSelectPipe,
                ],
                imports: [
                    CommonModule,
                ],
            },] },
];
/** @nocollapse */
ReduxModule.ctorParameters = () => [
    { type: Injector, },
    { type: ReduxReducerProvider, },
    { type: Array, decorators: [{ type: Optional }, { type: Inject, args: [ReduxStateDefinitionToken,] },] },
];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
const getActionType = ReduxActionDispatcher.getType;
const dispatch = ReduxActionDispatcher.dispatch;

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
/**
 * Generated bundle index. Do not edit.
 */

export { ReduxModule, getActionType, dispatch, ReduxSelect, ReduxReducerProvider, ReduxSelectPipe, ReduxSelector, ReduxStateDefinitionToken, ReduxStateProvider, ReduxStore, ReduxTestingStore, TestingStateProvider, ReduxRegistry };
//# sourceMappingURL=harmowatch-ngx-redux-core.js.map
